# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20231204

import logging
import os
import time
from threading import Thread
from typing import Union

import werkzeug
from flask import Response, abort, json, redirect, request, send_from_directory

import neetbox
from neetbox._protocol import *
from neetbox.server._bridge import Bridge

from .db import QueryCondition

werkzeug_log = logging.getLogger("werkzeug")
werkzeug_log.setLevel(logging.ERROR)  # disable flask http call logs


def get_flask_server(debug=False):
    __PROC_NAME = "NEETBOX"
    from neetbox.logging import LogStyle
    from neetbox.logging.logger import Logger, LogLevel

    logger = Logger("NEETBOX", LogStyle(skip_writers=["ws"]))

    if debug:
        logger.set_log_level(LogLevel.DEBUG)
        logger.log(f"Running with debug, using APIFlask")
        from apiflask import APIFlask

        app = APIFlask(__PROC_NAME, static_folder=None)
    else:
        logger.log(f"Running in production mode, using Flask")
        from flask import Flask

        app = Flask(__PROC_NAME, static_folder=None)

    front_end_dist_path = os.path.join(os.path.dirname(neetbox.__file__), "frontend_dist")

    @app.route("/")
    def static_serve_root():
        return redirect("/web/")

    @app.route("/web/")
    @app.route("/web/<path:name>")
    def static_serve(name=""):
        try:
            return send_from_directory(front_end_dist_path, name)
        except werkzeug.exceptions.NotFound:
            return send_from_directory(front_end_dist_path, "index.html")

    @app.route("/hello", methods=["GET"])
    def just_send_hello():
        return {"hello": "hello"}

    @app.route(f"{FRONTEND_API_ROOT}/list", methods=["GET"])
    def get_status_of_all_proejcts():
        return [_project_status_from_bridge(bridge) for _, bridge in Bridge.items()]

    @app.route(f"{FRONTEND_API_ROOT}/project/<project_id>", methods=["GET"])
    def get_status_of_project_id(project_id: str):
        if not Bridge.has(project_id):
            logger.debug(f"visiting none existing project id {project_id}", prefix="[404]")
            abort(404, {ERROR_KEY: "project id not found"})
        bridge = Bridge.of_id(project_id)
        return _project_status_from_bridge(bridge)

    def _project_status_from_bridge(bridge: Bridge):
        run_id_info_list = bridge.get_run_ids()
        name_of_project = None
        for run_id_info in reversed(run_id_info_list):
            config = bridge.get_status(run_id=run_id_info[RUN_ID_KEY], series="config")
            if NAME_KEY in config:
                name_of_project = config[NAME_KEY]
                break
        return {
            PROJECT_ID_KEY: bridge.project_id,
            "online": bridge.is_online(),
            NAME_KEY: name_of_project,
            "runids": run_id_info_list,
        }

    def get_history_json_of(project_id: str, table_name: str, condition=Union[dict, str]):
        if not Bridge.has(project_id):
            logger.debug(f"visiting none existing project id {project_id}", prefix="[404]")
            abort(404, {ERROR_KEY: "project id not found"})
        try:
            condition = QueryCondition.from_json(
                json.loads(condition) if isinstance(condition, str) else condition
            )
        except Exception as e:  # if failed to parse
            logger.debug(f"failed to parse condition string", prefix="400")
            abort(400, {ERROR_KEY: "failed to parse condition string"})
        return Bridge.of_id(project_id).read_json_from_history(
            table_name=table_name, condition=condition
        )

    @app.route(f"{FRONTEND_API_ROOT}/project/<project_id>/log", methods=["GET"])
    def get_history_log_of(project_id):
        return get_history_json_of(
            project_id=project_id,
            table_name=LOG_TABLE_NAME,
            condition=request.args.get("condition"),
        )

    @app.route(f"{FRONTEND_API_ROOT}/project/<project_id>/hardware", methods=["GET"])
    def get_history_hardware_of(project_id):
        return get_history_json_of(
            project_id=project_id,
            table_name=EVENT_TYPE_NAME_HARDWARE,
            condition=request.args.get("condition"),
        )

    @app.route(f"{FRONTEND_API_ROOT}/project/<project_id>/series/<table_name>", methods=["GET"])
    def get_series_list_of(project_id: str, table_name: str):  # client side function
        if not Bridge.has(project_id):
            logger.debug(f"visiting none existing project id {project_id}", prefix="[404]")
            abort(404, {ERROR_KEY: "project id not found"})
        result = Bridge.of_id(project_id).get_series_of(
            table_name, run_id=request.args.get(RUN_ID_KEY)
        )
        return result

    @app.route(f"{FRONTEND_API_ROOT}/project/<project_id>/run/<run_id>", methods=["GET"])
    def get_status_of(project_id, run_id):
        if not Bridge.has(project_id):
            logger.debug(f"visiting none existing project id {project_id}", prefix="[404]")
            abort(404, {ERROR_KEY: "project id not found"})
        result = Bridge.of_id(project_id).get_status(run_id=run_id)
        return result

    @app.route(f"{FRONTEND_API_ROOT}/project/<project_id>/run/<run_id>", methods=["PUT"])
    def set_get_metadata_of_run_id(project_id: str, run_id: str):
        new_metadata = request.json
        bridge = Bridge.of_id(project_id)
        old_metadata = bridge.historyDB.fetch_metadata_of_run_id(run_id=run_id)  # get old metadata
        for k, v in new_metadata.items():
            old_metadata[k] = v
        return bridge.historyDB.fetch_metadata_of_run_id(run_id=run_id, metadata=new_metadata)

    @app.route(f"{FRONTEND_API_ROOT}/project/<project_id>/run/<run_id>", methods=["DELETE"])
    def delete_run_id(project_id: str, run_id: str):
        bridge = Bridge.of_id(project_id)
        if bridge.is_online(run_id):  # cannot delete running projects
            abort(400, {ERROR_KEY: "can only delete history run id."})
        bridge.historyDB.delete_run_id(run_id)
        if 0 == len(bridge.get_run_ids()):  # check if all the run ids are deleted
            del Bridge._id2bridge[project_id]  # delete the empty bridge
        return {RESULT_KEY: "success"}

    @app.route(f"/image/<project_id>", methods=["POST"])
    def upload_image(project_id):
        if not Bridge.has(project_id):
            logger.debug(f"visiting none existing project id {project_id}", prefix="[404]")
            abort(404, {ERROR_KEY: "project id not found"})
        message = EventMsg.loads(request.form["json"])
        image_bytes = request.files["image"].read()
        message.id = Bridge.of_id(project_id).save_blob_to_history(
            table_name="image",
            run_id=message.run_id,
            series=message.series,
            meta_data=message.payload,
            timestamp=message.timestamp,
            blob_data=image_bytes,
            num_row_limit=message.history_len,
        )
        message.payload = message.payload or {}
        Bridge.of_id(project_id).ws_send_to_frontends(message)
        return {"result": "ok", "id": message.id}

    @app.route(f"{FRONTEND_API_ROOT}/project/<project_id>/image/<image_id>", methods=["GET"])
    def get_image_of(project_id, image_id: int):
        if not Bridge.has(project_id):
            logger.debug(f"visiting none existing project id {project_id}", prefix="[404]")
            abort(404, {ERROR_KEY: "project id not found"})
        meta = request.args.get("meta") is not None

        [(_, _, meta_data, image)] = Bridge.of_id(project_id).read_blob_from_history(
            table_name="image", condition=QueryCondition(id=image_id), meta_only=meta
        )

        return (
            Response(meta_data, mimetype="application/json")
            if meta
            else Response(image, mimetype="image")
        )

    @app.route(f"{FRONTEND_API_ROOT}/project/<project_id>/image", methods=["GET"])
    def get_history_image_metadata_of(project_id):
        if not Bridge.has(project_id):
            logger.debug(f"visiting none existing project id {project_id}", prefix="[404]")
            abort(404, {ERROR_KEY: "project id not found"})
        _json_data = json.loads(request.args.get("condition", default="{}"))
        try:
            condition = QueryCondition.from_json(_json_data)
        except Exception as e:  # if failed to parse
            return abort(400)
        query_results = Bridge.of_id(project_id).read_blob_from_history(
            table_name="image", condition=condition, meta_only=True
        )  # todo ?
        result = [{"imageId": id, "metadata": meta_data} for id, _, meta_data in query_results]
        return result

    @app.route(f"{FRONTEND_API_ROOT}/project/<project_id>/scalar", methods=["GET"])
    def get_history_scalar_of(project_id):
        return get_history_json_of(
            project_id=project_id, table_name="scalar", condition=request.args.get("condition")
        )

    @app.route(f"{FRONTEND_API_ROOT}/project/<project_id>/progress", methods=["GET"])
    def get_history_progress_of(project_id):
        return get_history_json_of(
            project_id=project_id, table_name="progress", condition=request.args.get("condition")
        )

    @app.route(f"/shutdown", methods=["POST"])
    def shutdown():
        def __sleep_and_shutdown(secs=1):
            time.sleep(secs)
            os._exit(0)

        Thread(target=__sleep_and_shutdown).start()  # shutdown after 3 seconds
        logger.log(f"BYE.")
        return {RESULT_KEY: f"shutdown in 1 seconds."}

    return app



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20231204

from typing import Dict, Tuple

from rich import box
from rich.console import Console
from rich.table import Table

from neetbox._protocol import *


def get_web_socket_server(config, debug=False):
    from websocket_server import WebsocketServer

    from neetbox.logging import LogStyle
    from neetbox.logging.logger import Logger, LogLevel
    from neetbox.server._bridge import Bridge

    console = Console()
    logger = Logger("NEETBOX", LogStyle(skip_writers=["ws"]))
    if debug:
        logger.set_log_level(LogLevel.DEBUG)

    port = config["port"] + 1
    logger.info(f"creating web socket server on port {port}")
    ws_server = WebsocketServer(host="0.0.0.0", port=port)
    connected_clients: Dict(
        int, Tuple(str, str, IdentityType)
    ) = {}  # {cid:(project_id, run_id,type)} store connection only

    def who_is(client):
        if client["id"] not in connected_clients:
            raise RuntimeError(
                f"Could not determine who is {client}, it seems that this client has not done handshake."
            )
        project_id, run_id, identity_type = connected_clients[client["id"]]
        return project_id, run_id, identity_type

    def handle_ws_connect(client, server):
        logger.info(
            f"client(id={client['id']}) on {client['address']} connected. waiting for handshake..."
        )
        # do nothing before handshake message

    def handle_ws_disconnect(client, server):
        ws_client_id = client["id"]
        if ws_client_id not in connected_clients:
            return  # client disconnected before handshake, returning anyway
        project_id, run_id, who = connected_clients[ws_client_id]
        bridge = Bridge.of_id(project_id)
        if not bridge:
            return  # do nothing if bridge has been deleted
        if who == IdentityType.WEB:  # is web ws, remove from bridge's web ws list
            _new_web_ws_list = [
                c for c in Bridge.of_id(project_id).web_ws_list if c["id"] != ws_client_id
            ]
            bridge.web_ws_list = _new_web_ws_list
        elif (
            who == IdentityType.CLI
        ):  # is cli ws, remove the run id from bridge's connected clients
            del bridge.cli_ws_dict[run_id]
        del connected_clients[ws_client_id]  # remove from connected clients records
        logger.info(
            f"a {who}(ws client id {ws_client_id}) disconnected from project '{project_id}'"
        )

    def on_event_type_handshake(client, server, message: EventMsg):
        ws_client_id = client["id"]
        if ws_client_id in connected_clients:  # client cannot handshake again
            logger.warn(
                f"websocket client(id={ws_client_id}) issued a duplicated handshake, ignoring..."
            )
            server.send_message(
                client=client,
                msg=EventMsg.merge(
                    message,
                    {
                        PAYLOAD_KEY: {ERROR_KEY: 400, REASON_KEY: "name not found"},
                        WHO_KEY: IdentityType.SERVER,
                    },
                ).dumps(),
            )
            return  # client cannot handshake again
        # assign this client to a Bridge
        console.log(f"handling handshake for {message.who} with project id '{message.project_id}'")

        if message.who == IdentityType.WEB:
            if not Bridge.has(message.project_id):  # there is no such bridge
                merge_msg = {
                    PAYLOAD_KEY: {ERROR_KEY: 404, REASON_KEY: "project id not found"},
                    WHO_KEY: IdentityType.SERVER,
                }  # no such bridge, dropping...

            else:  # new connection from frontend
                bridge = Bridge.of_id(message.project_id)
                bridge.web_ws_list.append(client)
                connected_clients[ws_client_id] = (
                    message.project_id,
                    None,
                    IdentityType.WEB,
                )  # web have None run id
                merge_msg = {
                    PAYLOAD_KEY: {RESULT_KEY: 200, REASON_KEY: "join success"},
                    WHO_KEY: IdentityType.SERVER,
                }
        elif message.who == IdentityType.CLI:
            # new connection from cli
            bridge = Bridge(project_id=message.project_id)
            target_run_id = message.run_id
            if target_run_id not in bridge.cli_ws_dict:
                bridge.cli_ws_dict[target_run_id] = client  # assign cli to bridge
                connected_clients[ws_client_id] = (
                    message.project_id,
                    message.run_id,
                    IdentityType.CLI,
                )
                merge_msg = {
                    PAYLOAD_KEY: {RESULT_KEY: 200, REASON_KEY: "join success"},
                    WHO_KEY: IdentityType.SERVER,
                }  # handshake 200
            else:  # run id already exist
                merge_msg = {
                    PAYLOAD_KEY: {
                        ERROR_KEY: 400,
                        REASON_KEY: "client with run id already connected",
                    },
                    WHO_KEY: IdentityType.SERVER,
                }
        else:
            logger.warn(f"unknown type({message.who}) of websocket trying to handshake")
            return  # unknown connection type, dropping...

        server.send_message(  # send handshake feedback to the client
            client=client,
            msg=EventMsg.merge(message, merge_msg).dumps(),
        )

        logger.ok(f"client(id={client['id']}) on {client['address']} handshake succeed.")

        if debug:
            table = Table(
                title="Connected Websockets", box=box.MINIMAL_DOUBLE_HEAD, show_lines=True
            )
            table.add_column(PROJECT_ID_KEY, justify="center", style="magenta", no_wrap=True)
            table.add_column(WHO_KEY, justify="center", style="cyan")
            table.add_column(f"web:ws/cli:{RUN_ID_KEY}", justify="center", style="green")
            for _, _bridge in Bridge.items():
                table.add_row(_bridge.project_id, "", "")
                for cli_run_id, _ in _bridge.cli_ws_dict.items():
                    table.add_row("", IdentityType.CLI, cli_run_id)
                for ws_web in _bridge.web_ws_list:
                    table.add_row(
                        "", IdentityType.WEB, f"websocket id {ws_web['id']} on {ws_web['address']}"
                    )
            console.print(table)
        return

    def on_event_type_json(
        message: EventMsg,
        forward_to: IdentityType = IdentityType.OTHERS,
        save_history=True,
    ):
        if not Bridge.has(message.project_id):
            # project id must exist
            # drop anyway if not exist
            logger.warn(f"handle {message.event_type}. {message.project_id} not found.")
            return
        bridge = Bridge.of_id(message.project_id)
        if save_history:
            message.id = bridge.save_json_to_history(
                table_name=message.event_type,
                json_data=message.payload,
                series=message.series,
                run_id=message.run_id,
                timestamp=message.timestamp,
                num_row_limit=message.history_len,
            )
        if forward_to:
            if forward_to == IdentityType.SELF:
                forward_to = message.who
            if forward_to == IdentityType.OTHERS:
                forward_to = (
                    IdentityType.WEB if message.who == IdentityType.CLI else IdentityType.CLI
                )
            if forward_to in [IdentityType.WEB, IdentityType.BOTH]:
                bridge.ws_send_to_frontends(message)  # forward to frontends
            elif forward_to in [IdentityType.CLI, IdentityType.BOTH]:
                bridge.ws_send_to_client(message, run_id=message.run_id)  # forward to frontends

        return  # return after handling log forwarding

    def on_event_type_status(message: EventMsg):
        bridge = Bridge.of_id(message.project_id)
        bridge.set_status(run_id=message.run_id, series=message.series, value=message.payload)

    def on_event_type_hyperparams(message: EventMsg):
        bridge = Bridge.of_id(message.project_id)
        current_hyperparams = bridge.get_status(
            run_id=message.run_id, series=EVENT_TYPE_NAME_HPARAMS
        )  # get hyper params from status
        if message.series:  # if series of hyperparams specified
            current_hyperparams[message.series] = message.payload
        else:
            for k, v in message.payload.items():
                current_hyperparams[k] = v
        bridge.set_status(
            run_id=message.run_id, series=EVENT_TYPE_NAME_HPARAMS, value=current_hyperparams
        )

    def handle_ws_message(client, server: WebsocketServer, message):
        try:
            message = EventMsg.loads(message)
        except Exception as e:
            logger.err(
                f"Illegal message format from client {client}: {message}, failed to parse cause {e}, dropping..."
            )
            return
        # handle event-type
        if message.event_type == EVENT_TYPE_NAME_HANDSHAKE:  # handle handshake
            on_event_type_handshake(client=client, server=server, message=message)
            return  # return after handling handshake
        try:
            _, _, identity_type = who_is(client)
        except:
            logger.debug(f"ws sending message without handshake, dropping...")
            return  # !!! not handling messages from cli/web without handshake. handshake is aspecial   case and should be handled anyway before this check.
        if not message.who:
            message.who = identity_type
        if message.who != identity_type:
            logger.warn(f"Illegal IdentityType: expect {identity_type} but got {message.who}")
            return  # !!! security check, who should match who
        else:  # handle regular event types
            if message.event_type == EVENT_TYPE_NAME_HPARAMS:
                on_event_type_hyperparams(message)
            elif message.event_type == EVENT_TYPE_NAME_STATUS:
                on_event_type_status(message)
            elif message.event_type == EVENT_TYPE_NAME_ACTION:
                on_event_type_json(
                    message=message, forward_to=IdentityType.OTHERS, save_history=False
                )
            else:
                on_event_type_json(
                    message=message,
                    forward_to=IdentityType.OTHERS,
                    save_history=True,
                )

    ws_server.set_fn_new_client(handle_ws_connect)
    ws_server.set_fn_client_left(handle_ws_disconnect)
    ws_server.set_fn_message_received(handle_ws_message)

    return ws_server



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20231013

import json
import os

import click
from rich.console import Console
from rich.table import Table

import neetbox.config._global as global_config
from neetbox._protocol import VERSION
from neetbox.client._client_web_apis import *
from neetbox.config._workspace import (
    _get_module_level_config,
    _init_workspace,
    _load_workspace_config,
)
from neetbox.logging.formatting import LogStyle
from neetbox.logging.logger import Logger
from neetbox.server._server import server_process
from neetbox.utils.massive import check_read_toml

console = Console()

logger = Logger("NEETBOX", style=LogStyle(with_datetime=False, skip_writers=["ws", "file"]))


def get_connection_config():
    return _get_module_level_config("client")


# def get_base_addr(port=0):
#     daemon_config = get_daemon_config()
#     _port = port or daemon_config["port"]
#     daemon_address = f"{daemon_config['host']}:{_port}"
#     return f"http://{daemon_address}/"


@click.group()
@click.option(
    "--verbose",
    "-v",
    help="set verbose flag to True",
    default=False,
)
@click.pass_context
def main(ctx, verbose: bool):
    ctx.ensure_object(dict)
    """This is NEETBOX cli"""
    ctx.obj["verbose"] = verbose


def _try_load_workspace_if_applicable():
    CONFIG_FILE_NAME = f"neetbox.toml"
    is_workspace = check_read_toml(CONFIG_FILE_NAME)
    if is_workspace:
        _load_workspace_config(load_only=True)


@main.command(name="version")
def version_command():
    print(VERSION)


@main.command(name="list")
def list_command():
    """Show list of connected project names"""
    _try_load_workspace_if_applicable()
    try:
        _response = get_list()
        table = Table(title="Running NEETBOX Projects")
        table.add_column(NAME_KEY, justify="center", style="magenta", no_wrap=True)
        table.add_column("online", justify="center", style="cyan")
        table.add_column("project id", justify="center", no_wrap=True)
        table.add_column("runs", justify="center", style="green")

        for pjt in _response:
            table.add_row(
                pjt[NAME_KEY],
                str(pjt["online"]),
                pjt[PROJECT_ID_KEY],
                "\n".join(pjt[STATUS_TABLE_NAME].keys()),
            )

        console.print(table)

        if not len(_response):
            console.print("*There is project no server")

    except Exception as e:  # noqa
        logger.log("Could not fetch data. Is there any project with NEETBOX running?")
        raise e


@main.command(name="status")
@click.argument("name", metavar="name", required=True)
def status_command(name):
    """Show running project status of given name"""
    _try_load_workspace_if_applicable()
    _response = None
    try:
        _response = get_status_of(project_id=name)
        click.echo(json.dumps(_response))
    except Exception as e:  # noqa
        logger.log("Could not fetch data. Is there any project with NEETBOX running?")


@main.command(name="serve")
@click.option(
    "--port", "-p", help="specify which port to launch", metavar="port", required=False, default=0
)
@click.option("--debug", "-d", is_flag=True, help="Run with debug mode", default=False)
def serve(port, debug):
    """serve neetbox server in attached mode"""
    _try_load_workspace_if_applicable()
    _daemon_config = get_connection_config()
    try:
        logger.log(f"Launching server using config: {_daemon_config}")
        if port:
            _daemon_config["port"] = port
        server_process(cfg=_daemon_config, debug=debug)
    except Exception as e:
        logger.err(f"Failed to launch a neetbox server: {e}")
        os._exit(-1)


@main.command(name="shutdown")
@click.option(
    "--port",
    "-p",
    help="specify which port to send shutdown",
    metavar="port",
    required=False,
    default=0,
)
def shutdown_server(port):
    """shutdown neetbox server on specific port"""
    _try_load_workspace_if_applicable()
    try:
        daemon_config = get_connection_config()
        if port:
            daemon_config["port"] = port
        _response = shutdown()
        click.echo(_response)
    except Exception as e:
        logger.err(
            f"Failed to shutdown the neetbox server: {e}. Is the server running on port {daemon_config['port']}?"
        )


@main.command()
@click.option("--name", "-n", help="set project name", metavar="name", required=False)
def init(name: str):
    """initialize current folder as workspace and generate the config file from defaults"""
    try:
        init_succeed = _init_workspace(name=name)
        if init_succeed:
            logger.console_banner("neetbox", font="ansishadow")
            logger.log("Welcome to NEETBOX")
    except Exception as e:
        logger.err(f"Failed to init here: {e}")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20230413

import inspect
import io
import json
import os
import pathlib
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from datetime import date, datetime
from typing import Any, Callable, Iterable, Optional, Union

from rich.console import Console

from neetbox.config import get_project_id, get_run_id
from neetbox.logging.formatting import LogStyle, colored_text, styled_text
from neetbox.utils import formatting
from neetbox.utils.framing import TracebackIdentity


class LogWriter(ABC):
    @abstractmethod
    def write(self, raw_log):
        pass


# ================== DEFINE LOG TYPE =====================


@dataclass
class RawLog:
    rich_msg: str
    style: LogStyle
    caller_identity: TracebackIdentity
    whom: Any = None
    prefix: Optional[str] = None
    datetime_format: Optional[str] = None
    with_identifier: Optional[bool] = None
    with_datetime: Optional[bool] = None
    skip_writers: Optional[list[str]] = None

    def write_by(self, writer: LogWriter) -> bool:
        skip_writers = (self.style.skip_writers or []) + (self.skip_writers or [])
        for swr in skip_writers:
            if isinstance(writer, RawLog.name2writerType[swr]):
                return False  # skip this writer, do not write
        writer.write(self)
        return False

    @property
    def json(self) -> dict:
        default_style = self.style
        # prefix
        prefix = self.prefix or default_style.prefix
        # composing datetime
        with_datetime = (
            default_style.with_datetime if self.with_datetime is None else self.with_datetime
        )
        timestamp = ""
        if with_datetime:
            datetime_fmt = self.datetime_format or default_style.datetime_format
            timestamp = datetime.now().strftime(datetime_fmt)

        # composing identifier
        whom = ""
        with_identifier = (
            default_style.with_identifier if self.with_identifier is None else self.with_identifier
        )
        if with_identifier:
            identity_seq = self.caller_identity.as_str_sequence(lineno=False)
            if len(identity_seq) > 1:
                identity_seq = identity_seq[1:]
            if len(identity_seq) > self.style.trace_level:
                identity_seq = identity_seq[: self.style.trace_level - 1]
            whom = (
                str(self.whom)
                if self.whom is not None
                else default_style.split_char_identity.join(identity_seq)
            )  # check identity
        return {"series": prefix, "timestamp": timestamp, "whom": whom, "message": self.rich_msg}

    def __repr__(self) -> str:
        return json.dumps(self.json, default=str)


# ================== CONSOLE LOG WRITER =====================


class DefaultDictThatReturnsKeyOnMissing(defaultdict):
    def __missing__(self, key):
        return key


_console_prefix_2_colored_text = DefaultDictThatReturnsKeyOnMissing(
    str,
    {
        "ok": colored_text("[ok]", "green"),
        "debug": colored_text("[debug]", "cyan"),
        "info": colored_text("[info]", "white"),
        "warning": colored_text("[warning]", "yellow"),
        "mention": colored_text("[mention]", "yellow"),
        "error": colored_text("[error]", "red"),
    },
)


class __ConsoleLogWriter(LogWriter):
    _console = Console()

    def write(self, raw_log: RawLog):
        _msg_dict = raw_log.json
        _style = raw_log.style
        _prefix = _msg_dict["series"]
        _prefix = _console_prefix_2_colored_text[_prefix] + " " if _prefix else _prefix
        rich_msg = str(
            _prefix
            + _msg_dict["timestamp"]
            + _style.split_char_cmd * min(len(_msg_dict["timestamp"]), 1)
            + styled_text(_msg_dict["whom"], style=_style)
            + _style.split_char_cmd * min(len(_msg_dict["whom"]), 1)
            + _msg_dict["message"]
        )
        self.__class__._console.print(rich_msg)


# console writer singleton
consoleLogWriter = __ConsoleLogWriter()


# ================== FILE LOG WRITER =====================


class FileLogWriter(LogWriter):
    # class level static pool
    PATH_2_FILE_WRITER = {}

    # instance level non-static things
    file_writer = None  # assign in __init__

    def __new__(cls, path, *args, **kwargs):
        # per file, per writer.
        file_abs_path = os.path.abspath(path)
        if os.path.isdir(file_abs_path):
            raise Exception("Target path is not a file.")
        filename = formatting.legal_file_name_of(os.path.basename(path))
        dirname = os.path.dirname(path) if len(os.path.dirname(path)) != 0 else "."
        if not os.path.exists(dirname):
            raise Exception(f"Could not find dictionary {dirname}")
        real_path = os.path.join(dirname, filename)
        if file_abs_path not in FileLogWriter.PATH_2_FILE_WRITER:
            newWriter = LogWriter.__new__(cls)
            newWriter.file_writer = open(real_path, "a", encoding="utf-8", buffering=1)
            FileLogWriter.PATH_2_FILE_WRITER[file_abs_path] = newWriter

        return FileLogWriter.PATH_2_FILE_WRITER[file_abs_path]

    def __init__(self, path) -> None:
        self.file_writer = open(path, "a", encoding="utf-8", buffering=1)

    def write(self, raw_log: RawLog):
        _msg_dict = raw_log.json
        _style = raw_log.style
        _series_text = f"[{_msg_dict['series']}]" if _msg_dict["series"] else ""
        text_msg = _style.split_char_txt.join(
            [_msg_dict["timestamp"], _series_text, _msg_dict["whom"], _msg_dict["message"]]
        )
        self.file_writer.write(text_msg + "\n")


# ================== WS LOG WRITER =====================


class _WebSocketLogWriter(LogWriter):
    # class level statics
    connection = None  # connection should be assigned by neetbox._daemon.client._client to avoid recursive import

    def write(self, raw_log: RawLog):
        json_data = raw_log.json
        payload = {"whom": json_data["whom"], "message": json_data["message"]}
        if _WebSocketLogWriter.connection:
            _WebSocketLogWriter.connection.ws_send(
                event_type="log",
                series=json_data["series"],
                payload=payload,
                timestamp=json_data["timestamp"],
            )


def _assign_connection_to_WebSocketLogWriter(conn):
    _WebSocketLogWriter.connection = conn


webSocketLogWriter = _WebSocketLogWriter()


# ================== JSON LOG WRITER =====================


class JsonLogWriter(FileLogWriter):
    def write(self, raw_log: RawLog):
        # todo convert to json and write to file
        pass


# ================== POST INIT TYPE REF =====================

RawLog.name2writerType = {
    "stdout": __ConsoleLogWriter,
    "file": FileLogWriter,
    "ws": _WebSocketLogWriter,
    "json": JsonLogWriter,
}



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20230414


from neetbox._protocol import *

from ._client import connection


def _get(api, root=None):
    r = connection.get(api=api, root=root)
    _data = r.json()
    return _data


def _post(api, root=None, data=None):
    r = connection.post(api=api, root=root, json=data)
    return r


def get_list(root=None):
    api = f"{FRONTEND_API_ROOT}/list"
    return _get(api, root=root)


def get_status_of(project_id, root=None):
    id2name = get_list(root=root)
    name2id = {v: k for k, v in id2name}  # todo resolve dup name
    if project_id not in name2id:
        return None
    _id = name2id[project_id]
    api = f"{FRONTEND_API_ROOT}/status/{_id}"
    return _get(api)


def shutdown(root=None):
    api = "/shutdown"
    return _post(api, root=root)



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20230414

import setproctitle

from neetbox._protocol import *
from neetbox.server._bridge import Bridge
from neetbox.server._flask_server import get_flask_server
from neetbox.server._websocket_server import get_web_socket_server


def server_process(cfg, debug=False):
    setproctitle.setproctitle("NEETBOX SERVER")
    from neetbox.logging import LogStyle
    from neetbox.logging.logger import Logger

    logger = Logger("NEETBOX", LogStyle(skip_writers=["ws"]))
    # load bridges
    Bridge.load_histories()  # load history files

    # websocket server
    ws_server = get_web_socket_server(config=cfg, debug=debug)
    Bridge._ws_server = ws_server  # add websocket server to bridge

    # http server
    flask_server = get_flask_server(debug=debug)

    # launch
    logger.log(f"launching websocket server...")
    ws_server.run_forever(threaded=True)

    port = cfg["port"]
    logger.log(f"launching flask server on port {port}")
    flask_server.run(host="0.0.0.0", port=port)


if __name__ == "__main__":
    server_process(
        {
            "port": 5000,
        },
        debug=True,
    )



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20231201

from . import _history as history
from ._condition import DbQueryFetchType, DbQuerySortType, QueryCondition

__all__ = [
    "history",
    "DbQueryFetchType",
    "DbQuerySortType",
    "QueryCondition",
]



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20231203

import os
from importlib.metadata import version

import toml

CONFIG_FILE_NAME = f"cli.neetbox.toml"
NEETBOX_VERSION = version("neetbox")
from uuid import uuid4

from neetbox._protocol import MACHINE_ID_KEY
from neetbox.utils.massive import (
    check_read_toml,
    get_user_config_directory,
    update_dict_recursively,
)

_GLOBAL_CONFIG = {
    "version": NEETBOX_VERSION,
    "servers": [{"address": "localhost", "port": 20202}],
    MACHINE_ID_KEY: uuid4(),
}

CONFIG_FILE_NAME = f"neetbox.toml"


def overwrite_create_local(config: dict):
    user_config_dir = get_user_config_directory()
    assert user_config_dir is not None
    neetbox_config_dir = os.path.join(user_config_dir, "neetbox")
    config_file_path = os.path.join(neetbox_config_dir, CONFIG_FILE_NAME)
    if not os.path.exists(config_file_path):  # config not exist, try to create
        if not os.path.exists(neetbox_config_dir):  # config folder not exist
            os.mkdir(neetbox_config_dir)
        assert os.path.isdir(neetbox_config_dir)
    with open(config_file_path, "w+") as config_file:
        toml.dump(config, config_file)


def read_create_local():
    global _GLOBAL_CONFIG
    user_config_dir = get_user_config_directory()
    assert user_config_dir is not None
    neetbox_config_dir = os.path.join(user_config_dir, "neetbox")
    config_file_path = os.path.join(neetbox_config_dir, CONFIG_FILE_NAME)
    if not os.path.exists(config_file_path):  # config not exist, try to create
        overwrite_create_local(_GLOBAL_CONFIG)
    # read local file
    user_cfg = check_read_toml(config_file_path)
    assert user_cfg
    for k, v in user_cfg.items():
        _GLOBAL_CONFIG[k] = v


def set(key, value):
    global _GLOBAL_CONFIG
    _GLOBAL_CONFIG[key] = value
    overwrite_create_local(_GLOBAL_CONFIG)


def get(key):
    read_create_local()
    return _GLOBAL_CONFIG.get(key, None)



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20230315

import functools
import os
from datetime import date
from enum import Enum
from random import randint
from typing import Any, Optional

from rich.console import Console
from rich.panel import Panel

from neetbox.config._workspace import (
    _get_module_level_config as get_module_level_config,
)
from neetbox.logging._writer import (
    FileLogWriter,
    JsonLogWriter,
    RawLog,
    consoleLogWriter,
    webSocketLogWriter,
)
from neetbox.logging.formatting import LogStyle
from neetbox.utils.framing import get_caller_identity_traceback


class LogLevel(Enum):
    ALL = 4
    DEBUG = 3
    INFO = 2
    WARNING = 1
    ERROR = 0

    def __lt__(self, other):
        return self.value < other.value

    def __le__(self, other):
        return self.value <= other.value

    def __eq__(self, other):
        return self.value == other.value

    def __ne__(self, other):
        return self.value != other.value

    def __gt__(self, other):
        return self.value > other.value

    def __ge__(self, other):
        return self.value >= other.value


class Logger:
    # global static
    __WHOM_2_LOGGER = {}
    __WHOM_2_STYLE = {}
    _console = Console()

    def __init__(self, whom: Any = None, style: Optional[LogStyle] = None):
        self.whom: Any = whom
        if style and style.console_color is None:
            style.randcolor()
        self.style: Optional[LogStyle] = style
        # default writing to console and ws
        self.console_writer = consoleLogWriter
        self.ws_writer = webSocketLogWriter
        self.file_writer = None
        _cfg = get_module_level_config()
        self.set_log_level(_cfg["level"])
        if style:
            if style.skip_writers and "file" not in style.skip_writers:
                self.set_log_dir(_cfg["logdir"])

    def __call__(self, whom: Any = None, style: Optional[LogStyle] = None) -> "Logger":
        """Attention! do not call this logger instance unless you know what are you doing. Users should use the default logger by import logger from neetbox.logging.

        Args:
            whom (Any, optional): who is logging, could be anything. Defaults to None(neetbox will automatically trace who is creating this logger if set to None).
            style (Optional[LogStyle], optional): overwrite default logger style. Defaults to None(neetbox will create a default style if set to None).

        Returns:
            Logger: a new logger
        """
        if whom is None:
            return DEFAULT_LOGGER
        if whom in Logger.__WHOM_2_LOGGER:
            return Logger.__WHOM_2_LOGGER[whom]
        Logger.__WHOM_2_LOGGER[whom] = Logger(whom=whom, style=style)
        return Logger.__WHOM_2_LOGGER[whom]

    def set_log_level(self, level: LogLevel):
        if type(level) is str:
            level = {
                "ALL": LogLevel.ALL,
                "DEBUG": LogLevel.DEBUG,
                "INFO": LogLevel.INFO,
                "WARNING": LogLevel.WARNING,
                "ERROR": LogLevel.ERROR,
            }[level]
        if type(level) is int:
            assert level >= 0 and level <= 3
            level = LogLevel(level)
        self.log_level = level

    def log(
        self,
        *content,
        prefix: Optional[str] = None,
        datetime_format: Optional[str] = None,
        with_identifier: Optional[bool] = None,
        with_datetime: Optional[bool] = None,
        skip_writers: list[str] = [],
        traceback=2,
    ):
        """log something

        Args:
            prefix (Optional[str], optional): prefix shows at the start of console log while it shows as a tag on frontend. Defaults to None.
            datetime_format (Optional[str], optional): change the format neetbox displays time. Defaults to None, neetbox will use default style ("%Y-%m-%dT%H:%M:%S.%f") if None is passed.
            with_identifier (Optional[bool], optional): whether to show who is logging, note that this option has noting todo with traceback. Defaults to None(True).
            with_datetime (Optional[bool], optional): whether to show datetime in logs. Defaults to None(True).
            skip_writers (list[str], optional): writers to skip, possible writes are 'stdout'(write into console), 'file'(write into file), 'ws'(write to frontend). Defaults to [], which means write to all writers.
            traceback (int, optional): level of traceback. Defaults to 2. Do not change this option unless you know what you are doing.

        Returns:
            Logger: returns itself that you can do chain function call
        """
        _caller_identity = get_caller_identity_traceback(traceback=traceback)

        # converting args into a single string
        _pure_str_message = ""
        for msg in content:
            _pure_str_message += str(msg) + " "

        if type(skip_writers) is str:
            skip_writers = [skip_writers]

        _style = self.style
        if not _style:  # if style not set
            _style_index = str(_caller_identity)
            if _style_index in Logger.__WHOM_2_STYLE:  # check for previous style
                _style = Logger.__WHOM_2_STYLE[_style_index]
            else:
                _style = LogStyle().randcolor()
                Logger.__WHOM_2_STYLE[_style_index] = _style

        raw_log = RawLog(
            rich_msg=_pure_str_message,
            caller_identity=_caller_identity,
            whom=self.whom,
            style=_style,
            prefix=prefix,
            datetime_format=datetime_format,
            with_identifier=with_identifier,
            with_datetime=with_datetime,
            skip_writers=skip_writers,
        )

        for writer in [self.console_writer, self.ws_writer, self.file_writer]:
            if writer:
                raw_log.write_by(writer)

        return self

    def ok(
        self,
        *content,
        datetime_format: Optional[str] = None,
        with_identifier: Optional[bool] = None,
        with_datetime: Optional[bool] = None,
        skip_writers: list[str] = [],
    ):
        if self.log_level >= LogLevel.INFO:
            self.log(
                *content,
                prefix=f"ok",
                skip_writers=skip_writers,
                traceback=3,
                datetime_format=datetime_format,
                with_identifier=with_identifier,
                with_datetime=with_datetime,
            )
        return self

    def debug(
        self,
        *content,
        prefix="debug",
        datetime_format: Optional[str] = None,
        with_identifier: Optional[bool] = None,
        with_datetime: Optional[bool] = None,
        skip_writers: list[str] = [],
    ):
        if self.log_level >= LogLevel.DEBUG:
            self.log(
                *content,
                prefix=prefix,
                skip_writers=skip_writers,
                traceback=3,
                datetime_format=datetime_format,
                with_identifier=with_identifier,
                with_datetime=with_datetime,
            )
        return self

    def info(
        self,
        *message,
        datetime_format: Optional[str] = None,
        with_identifier: Optional[bool] = None,
        with_datetime: Optional[bool] = None,
        skip_writers: list[str] = [],
    ):
        if self.log_level >= LogLevel.INFO:
            self.log(
                *message,
                prefix=f"info",
                skip_writers=skip_writers,
                traceback=3,
                datetime_format=datetime_format,
                with_identifier=with_identifier,
                with_datetime=with_datetime,
            )
        return self

    def warn(
        self,
        *message,
        datetime_format: Optional[str] = None,
        with_identifier: Optional[bool] = None,
        with_datetime: Optional[bool] = None,
        skip_writers: list[str] = [],
    ):
        if self.log_level >= LogLevel.WARNING:
            self.log(
                *message,
                prefix=f"warning",
                skip_writers=skip_writers,
                traceback=3,
                datetime_format=datetime_format,
                with_identifier=with_identifier,
                with_datetime=with_datetime,
            )
        return self

    def err(
        self,
        err,
        datetime_format: Optional[str] = None,
        with_identifier: Optional[bool] = None,
        with_datetime: Optional[bool] = None,
        skip_writers: list[str] = [],
        reraise=False,
    ):
        if self.log_level >= LogLevel.ERROR:
            self.log(
                str(err),
                prefix=f"error",
                skip_writers=skip_writers,
                traceback=3,
                datetime_format=datetime_format,
                with_identifier=with_identifier,
                with_datetime=with_datetime,
            )
        if reraise:
            raise err
        return self

    def mention(
        self,
        mention_args=True,
        mention_result=True,
        skip_writers=[],
        datetime_format=None,
        with_identifier=None,
        with_datetime=None,
    ):
        def with_logging(func):
            @functools.wraps(func)
            def _with_logging(*args, **kwargs):
                self.log(
                    f"Entering: {func.__name__}" + f", args={args}, kwargs={kwargs}"
                    if mention_args
                    else "",
                    prefix=f"mention",
                    skip_writers=skip_writers,
                    traceback=4,
                    datetime_format=datetime_format,
                    with_identifier=with_identifier,
                    with_datetime=with_datetime,
                )
                try:
                    result = func(*args, **kwargs)
                except Exception as e:
                    raise e
                self.log(
                    f"Leaving: {func.__name__}" + f", with returned value {result}"
                    if mention_result
                    else "",
                    prefix=f"mention",
                    skip_writers=skip_writers,
                    traceback=4,
                    datetime_format=datetime_format,
                    with_identifier=with_identifier,
                    with_datetime=with_datetime,
                )
                return result

            return _with_logging

        return with_logging

    def console_banner(self, text, font: Optional[str] = None):
        from pyfiglet import Figlet, FigletFont

        builtin_font_list = [
            "ansiregular",
            "ansishadow",
            "isometrixc2",
            "nscripts",
            "nvscript",
        ]
        if not font:
            font = builtin_font_list[randint(0, len(builtin_font_list)) - 1]

        if font not in FigletFont.getFonts():
            if font in builtin_font_list:  # builtin but not installed
                module_path = os.path.dirname(__file__)
                FigletFont.installFonts(f"{module_path}/flfs/{font}.flf")
            else:  # path?
                assert os.path.isfile(
                    font
                ), "The provided font is not a fontname or a font file path."
                file_name = os.path.basename(font)
                file = os.path.splitext(file_name)
                if file[0] not in FigletFont.getFonts():  # no installed file match the file name
                    try:
                        self.info(f"{file[0]} is not installed. Trying to install as a fontfile.")
                        FigletFont.installFonts(f"res/flfs/{font}.flf")
                    except Exception:
                        self.err("Could not install font {font}. Fallback to default.")
                        font = None
                else:
                    font = file[0]
        f = Figlet(font)
        rendered_text = f.renderText(text)
        Logger._console.print(Panel.fit(f"{rendered_text}", border_style="green"))
        return self

    def skip_lines(self, line_cnt=1, skip_writers: list[str] = []):
        """Let the logger log some empty lines

        Args:
            line_cnt (int, optional): how many empty line. Defaults to 1.

        Returns:
            _Logger : the logger instance itself
        """
        self.log(
            "\n" * line_cnt, with_datetime=False, with_identifier=False, skip_writers=skip_writers
        )
        return self

    def log_txt_file(self, file, skip_writers: list[str] = []):
        if isinstance(file, str):
            file = open(file)
        context = ""
        for line in file.readlines():
            context += line
        self.log(context, with_datetime=False, with_identifier=False, skip_writers=skip_writers)
        return self

    def set_log_dir(self, path, independent=False):
        if not path:
            self._bind_file(None)
            return self
        if os.path.isfile(path):
            raise Exception("Target path is not a directory.")
        if not os.path.exists(path):
            try:
                os.makedirs(path)
            except Exception:
                raise Exception(f"Failed when trying to create directory {path}")
        log_file_name = ""
        if independent:
            log_file_name += self.whom
        log_file_name += str(date.today()) + ".log"
        self._bind_file(os.path.join(path, log_file_name))
        return self

    def _bind_file(self, path):
        if not path:
            self.file_writer = None
            return self
        self.file_writer = FileLogWriter(path=path)
        return self


DEFAULT_LOGGER = Logger(None)
